@@ -40,6 +40,7 @@ from boltons.iterutils import remap
 from pymatgen.core import Structure as PmgStructure
 from concurrent.futures import as_completed
 from requests_futures.sessions import FuturesSession
+from urllib3.util.retry import Retry
 from filetype.types.archive import Gz
 from filetype.types.image import Jpeg, Png, Gif, Tiff
 from pint import UnitRegistry
@@ -48,6 +49,7 @@ from pint.converters import ScaleConverter
 from pint.errors import DimensionalityError
 from datetime import datetime
 
+RETRIES = 3
 MAX_WORKERS = 10
 MAX_ELEMS = 10
 MAX_BYTES = 200 * 1024
@@ -125,18 +127,6 @@ url_format = SwaggerFormat(
 )
 
 
-def chunks(lst, n):
-    if isinstance(lst, set):
-        lst = list(lst)
-    elif not isinstance(lst, list):
-        raise ValueError("chunks needs list or set as input")
-
-    n = max(1, n)
-    for i in range(0, len(lst), n):
-        to = i + n
-        yield lst[i:to]
-
-
 # https://stackoverflow.com/a/8991553
 def grouper(n, iterable):
     it = iter(iterable)
@@ -148,7 +138,14 @@ def grouper(n, iterable):
 
 
 def get_session():
-    s = FuturesSession(max_workers=MAX_WORKERS)
+    adapter_kwargs = dict(max_retries=Retry(
+        total=RETRIES,
+        read=RETRIES,
+        connect=RETRIES,
+        respect_retry_after_header=True,
+        status_forcelist=[429],  # rate limit
+    ))
+    s = FuturesSession(max_workers=MAX_WORKERS, adapter_kwargs=adapter_kwargs)
     s.hooks['response'].append(_response_hook)
     return s
 
@@ -342,8 +339,9 @@ def _run_futures(futures, total: int = 0, timeout: int = -1):
                     responses[future.track_id] = response.result
 
                 elapsed = time.perf_counter() - start
+                timed_out = timeout > 0 and elapsed > timeout
 
-                if timeout > 0 and elapsed > timeout:
+                if timed_out or not response.ok:
                     for future in futures:
                         future.cancel()
 
@@ -372,9 +370,9 @@ class Client(SwaggerClient):
         """Initialize the client - only reloads API spec from server as needed
 
         Args:
-            apikey (str): API key - can also be set via MPCONTRIBS_API_KEY env var
-            headers (dict): custom headers for localhost connections - ignored if API key set
-            host (str): host address to connect to - can also be set via MPCONTRIBS_API_HOST
+            apikey (str): API key (or use MPCONTRIBS_API_KEY env var) - ignored if headers set
+            headers (dict): custom headers for localhost connections
+            host (str): host address to connect to (or use MPCONTRIBS_API_HOST env var)
         """
         # - Kong forwards consumer headers when api-key used for auth
         # - forward consumer headers when connecting through localhost
@@ -398,6 +396,7 @@ class Client(SwaggerClient):
         ssl = host.endswith(".materialsproject.org") and not host.startswith("localhost.")
         self.protocol = "https" if ssl else "http"
         self.url = f"{self.protocol}://{self.host}"
+        self.session = get_session()
 
         if self.url not in VALID_URLS:
             raise ValueError(f"{self.url} not a valid URL (one of {VALID_URLS})")
@@ -409,6 +408,12 @@ class Client(SwaggerClient):
         ):
             self._load()
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        return self.session.close()
+
     def _load(self):
         http_client = FidoClientGlobalHeaders(headers=self.headers)
         loader = Loader(http_client)
@@ -451,7 +456,7 @@ class Client(SwaggerClient):
                         col = f"{col}__value"
                         columns["number"].append(col)
 
-        operators = {"text": ["contains"], "number": ["gte", "lte"]}
+        operators = {"text": ["exact", "contains"], "number": ["gte", "lte"]}
         for path, d in spec_dict["paths"].items():
             for verb in ["get", "put", "post", "delete"]:
                 if verb in d:
@@ -512,6 +517,70 @@ class Client(SwaggerClient):
         _, per_page_max = self._get_per_page_default_max(op=op, resource=resource)
         return min(per_page_max, per_page)
 
+    def _split_query(
+        self,
+        query: dict,
+        op: str = "get",
+        resource: str = "contributions",
+        pages: int = -1,
+    ) -> List[dict]:
+        """Avoid URI too long errors"""
+        pp_default, pp_max = self._get_per_page_default_max(op=op, resource=resource)
+        per_page = pp_default if "id__in" in query else pp_max
+        query["per_page"] = per_page
+        nr_params_to_split = sum(
+            len(v) > per_page for v in query.values() if isinstance(v, list)
+        )
+        if nr_params_to_split > 1:
+            raise NotImplementedError(
+                f"More than one list in query with length > {per_page} not supported!"
+            )
+
+        queries = []
+
+        for k, v in query.items():
+            if isinstance(v, list) and len(v) > per_page:
+                for chunk in grouper(per_page, v):
+                    queries.append({k: list(chunk)})
+
+        if not queries:
+            queries = [query]
+
+        if len(queries) == 1 and pages and pages > 0:
+            queries = []
+            for page in range(1, pages+1):
+                queries.append(deepcopy(query))
+                queries[-1]["page"] = page
+
+        for q in queries:
+            # copy over missing parameters
+            for k, v in query.items():
+                if k not in q:
+                    q[k] = v
+
+            # comma-separated lists
+            for k, v in q.items():
+                if isinstance(v, list):
+                    q[k] = ",".join(v)
+
+        return queries
+
+    def _get_future(
+        self,
+        track_id,
+        params: dict,
+        rel_url: str = "contributions",
+        op: str = "get"
+    ):
+        if self.session and self.session.executor._shutdown:
+            raise ValueError("Session closed. Use `with` statement.")
+
+        future = getattr(self.session, op)(
+            f"{self.url}/{rel_url}/", headers=self.headers, params=params
+        )
+        setattr(future, "track_id", track_id)
+        return future
+
     def get_project_names(self) -> List[str]:
         """Retrieve list of project names."""
         resp = self.projects.get_entries(_fields=["name"]).result()
@@ -762,9 +831,7 @@ class Client(SwaggerClient):
         self.projects.update_entry(pk=name, project={"columns": []}).result()
         return self.projects.update_entry(pk=name, project=payload).result()
 
-    def delete_contributions(
-            self, name: str, per_page: int = 100, retry: bool = False, timeout: int = -1
-    ):
+    def delete_contributions(self, name: str):
         """Remove all contributions for a project
 
         Note: This also resets the columns field for a project. It might have to be
@@ -772,73 +839,68 @@ class Client(SwaggerClient):
 
         Args:
             name (str): name of the project for which to delete contributions
-            per_page (int): number of contributions to delete per request
-            retry (bool): if True, retry deletion of failed contributions until done
-            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
         tic = time.perf_counter()
-        query = dict(project=name)
-        cids = self.get_all_ids(query=query, timeout=timeout).get(name, {}).get("ids", [])
-        total = len(cids)
         # reset columns to be save (sometimes not all are reset BUGFIX?)
         self.projects.update_entry(pk=name, project={"columns": []}).result()
-        per_page = self._get_per_page(per_page, op="delete")
+        cids = self.get_all_ids(dict(project=name)).get(name, {}).get("ids", [])
 
-        if cids:
-            with get_session() as session:
-                while cids:
-                    futures = [
-                        session.delete(
-                            f"{self.url}/contributions/",
-                            headers=self.headers,
-                            params={
-                                "project": name,
-                                "id__in": ",".join(chunk),
-                                "per_page": per_page,
-                            },
-                        )
-                        for chunk in chunks(cids, per_page)
-                    ]
-
-                    _run_futures(futures, total=len(cids), timeout=timeout)
-                    cids = self.get_all_ids(
-                        query=query, timeout=timeout
-                    ).get(name, {}).get("ids", [])
-
-                    if not retry:
-                        break
-
-                self._load()
-
-            toc = time.perf_counter()
-            dt = (toc - tic) / 60
-
-            if cids:
-                print(f"There were errors and {len(cids)} contributions are left to delete!")
-            else:
-                print(f"It took {dt:.1f}min to delete {total} contributions.")
-        else:
+        if not cids:
             print(f"There aren't any contributions to delete for {name}")
+            return
 
-    def get_totals(self, query: dict = None, per_page: int = 1) -> tuple:
-        """Retrieve total count and pages for contributions matching query
+        total = len(cids)
+        query = {"id__in": cids}
+        _, total_pages = self.get_totals(query=query)
+        queries = self._split_query(query, op="delete", pages=total_pages)
+        futures = [self._get_future(i, q, op="delete") for i, q in enumerate(queries)]
+        _run_futures(futures, total=total)
+        left, _ = self.get_totals(query=dict(project=name))
+        deleted = total - left
+        self._load()
+        toc = time.perf_counter()
+        dt = (toc - tic) / 60
+        print(f"It took {dt:.1f}min to delete {deleted} contributions.")
 
-        See `client.contributions.get_entries()` for query.
+        if left:
+            print(f"There were errors and {left} contributions are left to delete!")
+
+    def get_totals(
+        self,
+        query: dict = None,
+        timeout: int = -1,
+        resource: str = "contributions",
+        op: str = "get"
+    ) -> tuple:
+        """Retrieve total count and pages for resource entries matching query
 
         Args:
-            query (dict): query to select contributions
-            per_page (int): number of contributions per page to calculate correct #pages
+            query (dict): query to select resource entries
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
+            op (str): operation to calculate total pages for, one of
+                      ("get", "create", "update", "delete", "download")
 
         Returns:
             tuple of total counts and pages
         """
+        ops = {"get", "create", "update", "delete", "download"}
+        if op not in ops:
+            print(f"`op` has to be one of {ops}")
+            return
+
         query = query or {}
         skip_keys = {"per_page", "_fields", "format"}
         query = {k: v for k, v in query.items() if k not in skip_keys}
-        per_page = self._get_per_page(per_page)
-        result = self.contributions.get_entries(
-            per_page=per_page, _fields=["id"], **query
-        ).result()
+        query["_fields"] = ["_id"]
+        queries = self._split_query(query, resource=resource, op=op)  # don't paginate
+        result = {"total_count": 0, "total_pages": 0}
+        futures = [self._get_future(i, q, rel_url=resource) for i, q in enumerate(queries)]
+        responses = _run_futures(futures, timeout=timeout)
+
+        for resp in responses.values():
+            for k in result:
+                result[k] += resp[k]
+
         return result["total_count"], result["total_pages"]
 
     def get_unique_identifiers_flags(self, projects: list = None) -> dict:
@@ -869,6 +931,7 @@ class Client(SwaggerClient):
         timeout: int = -1,
         data_id_fields: dict = None,
         fmt: str = "sets",
+        op: str = "get",
     ) -> dict:
         """Retrieve a list of existing contribution and component (Object)IDs
 
@@ -878,6 +941,8 @@ class Client(SwaggerClient):
             timeout (int): cancel remaining requests if timeout exceeded (in seconds)
             data_id_fields (dict): map of project to extra field in `data` to include as ID field
             fmt (str): return `sets` of identifiers or `map` (see below)
+            op (str): operation to calculate total pages for, one of
+                      ("get", "create", "update", "delete", "download")
 
         Returns:
             {"<project-name>": {
@@ -925,6 +990,11 @@ class Client(SwaggerClient):
             print(f"`fmt` must be subset of {fmts}!")
             return
 
+        ops = {"get", "create", "update", "delete", "download"}
+        if op not in ops:
+            print(f"`op` has to be one of {ops}")
+            return
+
         unique_identifiers = self.get_unique_identifiers_flags()
         data_id_fields = {
             k: v for k, v in data_id_fields.items()
@@ -932,10 +1002,8 @@ class Client(SwaggerClient):
         } if data_id_fields else {}
 
         ret = {}
-        _, per_page = self._get_per_page_default_max()
         query = query or {}
         [query.pop(k, None) for k in ["page", "per_page", "_fields"]]
-        _, pages = self.get_totals(query=query, per_page=per_page)
         id_fields = {"project", "id", "identifier"}
 
         if data_id_fields:
@@ -944,93 +1012,72 @@ class Client(SwaggerClient):
                 for data_id_field in data_id_fields.values()
             )
 
-        fields = ",".join(id_fields | components)
-        url = f"{self.url}/contributions/"
+        query["_fields"] = list(id_fields | components)
+        _, total_pages = self.get_totals(query=query, timeout=timeout)
+        queries = self._split_query(query, op=op, pages=total_pages)
+        futures = [self._get_future(i, q) for i, q in enumerate(queries)]
+        responses = _run_futures(futures, timeout=timeout)
 
-        # convert lists in query to comma-separated
-        for k, v in query.items():
-            if isinstance(v, list):
-                query[k] = ",".join(v)
+        for resp in responses.values():
+            for contrib in resp["data"]:
+                project = contrib["project"]
+                data_id_field = data_id_fields.get(project)
 
-        with get_session() as session:
+                if fmt == "sets":
+                    if project not in ret:
+                        id_keys = ["ids", "identifiers"]
+                        if data_id_field:
+                            id_field = f"{data_id_field}_set"
+                            id_keys.append(id_field)
 
-            def get_future(page):
-                params = {"page": page, "per_page": per_page, "_fields": fields}
-                params.update(query)
-                future = session.get(url, headers=self.headers, params=params)
-                setattr(future, "track_id", page)
-                return future
+                        ret[project] = {k: set() for k in id_keys}
 
-            futures = [get_future(page + 1) for page in range(pages)]
+                    ret[project]["ids"].add(contrib["id"])
+                    ret[project]["identifiers"].add(contrib["identifier"])
 
-            while futures:
-                responses = _run_futures(futures, timeout=timeout)
+                    if data_id_field:
+                        ret[project][id_field].add(contrib["data"][data_id_field])
 
-                for resp in responses.values():
-                    for contrib in resp["data"]:
-                        project = contrib["project"]
-                        data_id_field = data_id_fields.get(project)
+                    for component in components:
+                        if component in contrib:
+                            if component not in ret[project]:
+                                ret[project][component] = {"ids": set(), "md5s": set()}
 
-                        if fmt == "sets":
-                            if project not in ret:
-                                id_keys = ["ids", "identifiers"]
-                                if data_id_field:
-                                    id_field = f"{data_id_field}_set"
-                                    id_keys.append(id_field)
+                            for d in contrib[component]:
+                                for k in ["id", "md5"]:
+                                    ret[project][component][f"{k}s"].add(d[k])
 
-                                ret[project] = {k: set() for k in id_keys}
+                elif fmt == "map":
+                    identifier = contrib["identifier"]
+                    data_id_field_val = contrib.get("data", {}).get(data_id_field)
 
-                            ret[project]["ids"].add(contrib["id"])
-                            ret[project]["identifiers"].add(contrib["identifier"])
+                    if project not in ret:
+                        ret[project] = {}
 
-                            if data_id_field:
-                                ret[project][id_field].add(contrib["data"][data_id_field])
+                    if unique_identifiers[project]:
+                        ret[project][identifier] = {"id": contrib["id"]}
 
-                            for component in components:
-                                if component in contrib:
-                                    if component not in ret[project]:
-                                        ret[project][component] = {"ids": set(), "md5s": set()}
+                        if data_id_field and data_id_field_val:
+                            ret[project][identifier][data_id_field] = data_id_field_val
 
-                                    for d in contrib[component]:
-                                        for k in ["id", "md5"]:
-                                            ret[project][component][f"{k}s"].add(d[k])
-
-                        elif fmt == "map":
-                            identifier = contrib["identifier"]
-                            data_id_field_val = contrib.get("data", {}).get(data_id_field)
-
-                            if project not in ret:
-                                ret[project] = {}
-
-                            if unique_identifiers[project]:
-                                ret[project][identifier] = {"id": contrib["id"]}
-
-                                if data_id_field and data_id_field_val:
-                                    ret[project][identifier][data_id_field] = data_id_field_val
-
-                                for component in components:
-                                    if component in contrib:
-                                        ret[project][identifier][component] = {
-                                            d["name"]: {"id": d["id"], "md5": d["md5"]}
-                                            for d in contrib[component]
-                                        }
-
-                            elif data_id_field and data_id_field_val:
-                                ret[project][identifier] = {
-                                    data_id_field_val: {"id": contrib["id"]}
+                        for component in components:
+                            if component in contrib:
+                                ret[project][identifier][component] = {
+                                    d["name"]: {"id": d["id"], "md5": d["md5"]}
+                                    for d in contrib[component]
                                 }
 
-                                for component in components:
-                                    if component in contrib:
-                                        ret[project][identifier][data_id_field_val][component] = {
-                                            d["name"]: {"id": d["id"], "md5": d["md5"]}
-                                            for d in contrib[component]
-                                        }
+                    elif data_id_field and data_id_field_val:
+                        ret[project][identifier] = {
+                            data_id_field_val: {"id": contrib["id"]}
+                        }
 
-                futures = [
-                    future for future in futures
-                    if not future.cancelled() and future.track_id not in responses.keys()
-                ]
+                        for component in components:
+                            if component in contrib:
+                                ret[project][identifier][data_id_field_val][component] = {
+                                    d["name"]: {"id": d["id"], "md5": d["md5"]}
+                                    for d in contrib[component]
+                                }
 
         return ret
 
@@ -1094,7 +1141,7 @@ class Client(SwaggerClient):
         contributions: List[dict],
         ignore_dupes: bool = False,
         retry: bool = False,
-        per_page: int = 100,
+        per_request: int = 100,
         timeout: int = -1
     ):
         """Submit a list of contributions
@@ -1123,7 +1170,7 @@ class Client(SwaggerClient):
             contributions (list): list of contribution dicts to submit
             ignore_dupes (bool): force duplicate components to be submitted
             retry (bool): keep trying until all contributions successfully submitted
-            per_page (int): number of contributions to submit in each chunk/request
+            per_request (int): number of contributions to submit per request
             timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
         if not contributions or not isinstance(contributions, list):
@@ -1135,7 +1182,7 @@ class Client(SwaggerClient):
         project_names = set()
         collect_ids = []
         require_one_of = {"data"} | set(COMPONENTS)
-        per_page = self._get_per_page(per_page)
+        per_page = self._get_per_page(per_request)
 
         for idx, c in enumerate(contributions):
             has_keys = require_one_of & c.keys()
@@ -1156,21 +1203,22 @@ class Client(SwaggerClient):
 
         id2project = {}
         if collect_ids:
-            resp = self.get_all_ids(
-                query=dict(id__in=collect_ids), timeout=timeout
-            )
+            resp = self.get_all_ids(dict(id__in=collect_ids), timeout=timeout)
             project_names |= set(resp.keys())
 
             for project_name, values in resp.items():
                 for cid in values["ids"]:
                     id2project[cid] = project_name
 
-        print("get existing contributions ...")
+        existing = defaultdict(dict)
         project_names = list(project_names)
-        unique_identifiers = self.get_unique_identifiers_flags(projects=project_names)
-        existing = defaultdict(dict, self.get_all_ids(
-            query=dict(project__in=project_names), include=COMPONENTS, timeout=timeout
-        ))
+
+        if len(collect_ids) != len(contributions):
+            print("get existing contributions ...")
+            unique_identifiers = self.get_unique_identifiers_flags(projects=project_names)
+            existing = defaultdict(dict, self.get_all_ids(
+                dict(project__in=project_names), include=COMPONENTS, timeout=timeout
+            ))
 
         # prepare contributions
         print("prepare contributions ...")
@@ -1308,70 +1356,71 @@ class Client(SwaggerClient):
 
         # submit contributions
         if contribs:
+            if self.session and self.session.executor._shutdown:
+                raise ValueError("Session closed. Use `with` statement.")
+
             print("submit contributions ...")
-            with get_session() as session:
-                total = 0
+            total = 0
 
-                def post_future(chunk):
-                    return session.post(
-                        f"{self.url}/contributions/",
-                        headers=self.headers,
-                        data=json.dumps(chunk).encode("utf-8"),
-                    )
+            def post_future(chunk):
+                return self.session.post(
+                    f"{self.url}/contributions/",
+                    headers=self.headers,
+                    data=json.dumps(chunk).encode("utf-8"),
+                )
 
-                def put_future(cdct):
-                    pk = cdct.pop("id")
-                    return session.put(
-                        f"{self.url}/contributions/{pk}/",
-                        headers=self.headers,
-                        data=json.dumps(cdct).encode("utf-8"),
-                    )
+            def put_future(cdct):
+                pk = cdct.pop("id")
+                return self.session.put(
+                    f"{self.url}/contributions/{pk}/",
+                    headers=self.headers,
+                    data=json.dumps(cdct).encode("utf-8"),
+                )
 
-                for project_name in project_names:
-                    ncontribs = len(contribs[project_name])
-                    total += ncontribs
-                    start = datetime.utcnow()
+            for project_name in project_names:
+                ncontribs = len(contribs[project_name])
+                total += ncontribs
+                start = datetime.utcnow()
 
-                    while contribs[project_name]:
-                        futures = []
-                        for chunk in chunks(contribs[project_name], per_page):
-                            post_chunk = []
-                            for c in chunk:
-                                if "id" in c:
-                                    futures.append(put_future(c))
-                                else:
-                                    post_chunk.append(c)
+                while contribs[project_name]:
+                    futures = []
+                    for chunk in grouper(per_page, contribs[project_name]):
+                        post_chunk = []
+                        for c in chunk:
+                            if "id" in c:
+                                futures.append(put_future(c))
+                            else:
+                                post_chunk.append(c)
 
-                            if post_chunk:
-                                futures.append(post_future(post_chunk))
+                        if post_chunk:
+                            futures.append(post_future(post_chunk))
 
-                        _run_futures(futures, total=ncontribs, timeout=timeout)
+                    _run_futures(futures, total=ncontribs, timeout=timeout)
 
-                        if unique_identifiers[project_name] and retry:
-                            existing[project_name] = self.get_all_ids(
-                                query=dict(project=project_name), include=COMPONENTS,
-                                timeout=timeout
-                            ).get(project_name, {"identifiers": set()})
-                            unique_identifiers[project_name] = self.projects.get_entry(
-                                pk=project_name, _fields=["unique_identifiers"]
-                            ).result()["unique_identifiers"]
-                            contribs[project_name] = [
-                                c for c in contribs[project_name]
-                                if c["identifier"] not in existing[project_name]["identifiers"]
-                            ]
-                        else:
-                            contribs[project_name] = []  # abort retrying
-                            if not unique_identifiers[project_name] and retry:
-                                print("Please resubmit failed contributions manually.")
+                    if retry and unique_identifiers[project_name]:
+                        existing[project_name] = self.get_all_ids(
+                            dict(project=project_name), include=COMPONENTS, timeout=timeout
+                        ).get(project_name, {"identifiers": set()})
+                        unique_identifiers[project_name] = self.projects.get_entry(
+                            pk=project_name, _fields=["unique_identifiers"]
+                        ).result()["unique_identifiers"]
+                        contribs[project_name] = [
+                            c for c in contribs[project_name]
+                            if c["identifier"] not in existing[project_name]["identifiers"]
+                        ]
+                    else:
+                        contribs[project_name] = []  # abort retrying
+                        if retry and not unique_identifiers[project_name]:
+                            print("Please resubmit failed contributions manually.")
 
-                end = datetime.utcnow()
-                updated_total, _ = self.get_totals(query=dict(
-                    project__in=project_names, last_modified__gt=start, last_modified__lt=end
-                ))
-                toc = time.perf_counter()
-                dt = (toc - tic) / 60
-                self._load()
-                print(f"It took {dt:.1f}min to submit {updated_total} contributions.")
+            end = datetime.utcnow()
+            updated_total, _ = self.get_totals(query=dict(
+                project__in=project_names, last_modified__gt=start, last_modified__lt=end
+            ), timeout=timeout)
+            toc = time.perf_counter()
+            dt = (toc - tic) / 60
+            self._load()
+            print(f"It took {dt:.1f}min to submit {updated_total} contributions.")
         else:
             print("Nothing to submit.")
 
@@ -1405,7 +1454,7 @@ class Client(SwaggerClient):
             print(f"`include` must be subset of {COMPONENTS}!")
             return
 
-        all_ids = self.get_all_ids(query=query, include=components, timeout=timeout)
+        all_ids = self.get_all_ids(query, include=components, timeout=timeout)
         fmt = query.get("format", "json")
         ndownloads = 0
 
@@ -1418,7 +1467,7 @@ class Client(SwaggerClient):
                 start = time.perf_counter()
 
             cids = list(values["ids"])
-            paths, per_page = self._download_resource(
+            paths = self._download_resource(
                 resource="contributions", ids=cids, fmt=fmt,
                 outdir=outdir, overwrite=overwrite, timeout=timeout
             )
@@ -1438,7 +1487,7 @@ class Client(SwaggerClient):
                     start = time.perf_counter()
 
                 ids = list(values[component]["ids"])
-                paths, per_page = self._download_resource(
+                paths = self._download_resource(
                     resource=component, ids=ids, fmt=fmt,
                     outdir=outdir, overwrite=overwrite, timeout=timeout
                 )
@@ -1558,40 +1607,33 @@ class Client(SwaggerClient):
             print(f"`fmt` must be one of {formats}!")
             return
 
-        oids = sorted(ObjectId(i) for i in ids if ObjectId.is_valid(i))
+        oids = sorted(i for i in ids if ObjectId.is_valid(i))
         outdir = Path(outdir) or Path(".")
         subdir = outdir / resource
         subdir.mkdir(parents=True, exist_ok=True)
         model = self.get_model(f"{resource.capitalize()}Schema")
         fields = list(model._properties.keys())
-        # NOTE use default per_page to avoid "URL too long" error
-        per_page, _ = self._get_per_page_default_max(op="download", resource=resource)
-        chunked_oids = grouper(per_page, map(str, oids))
+        query = {"format": fmt, "_fields": fields, "id__in": oids}
+        _, total_pages = self.get_totals(
+            query=query, resource=resource, op="download", timeout=timeout
+        )
+        queries = self._split_query(query, resource=resource, op="download", pages=total_pages)
         paths, futures = [], []
 
-        with get_session() as session:
-            for chunk in chunked_oids:
-                digest = get_md5({"ids": chunk})
-                path = subdir / f"{digest}.{fmt}.gz"
+        for query in queries:
+            digest = get_md5({"ids": query["id__in"].split(",")})
+            path = subdir / f"{digest}.{fmt}.gz"
 
-                if not path.exists() or overwrite:
-                    future = session.get(
-                        f"{self.url}/{resource}/download/gz/",
-                        headers=self.headers,
-                        params={
-                            "format": fmt,
-                            "id__in": ",".join(chunk),
-                            "_fields": ",".join(fields),
-                        },
-                    )
-                    setattr(future, "track_id", path)
-                    futures.append(future)
+            if not path.exists() or overwrite:
+                futures.append(self._get_future(
+                    path, query, rel_url=f"{resource}/download/gz"
+                ))
 
-            if futures:
-                responses = _run_futures(futures, timeout=timeout)
+        if futures:
+            responses = _run_futures(futures, timeout=timeout)
 
-                for path, resp in responses.items():
-                    path.write_bytes(resp)
-                    paths.append(path)
+            for path, resp in responses.items():
+                path.write_bytes(resp)
+                paths.append(path)
 
-        return paths, per_page
+        return paths
